'''
Created on 2023-08-16

@author: wf
'''
import sys
from argparse import ArgumentParser
from nicetrack.webserver import WebServer
from ngwidgets.cmd import WebserverCmd

class NiceTrackCmd(WebserverCmd):
    """
    nicetrack command line handling
    """
    
    def __init__(self):
        """
        constructor
        """
        config=WebServer.get_config()
        WebserverCmd.__init__(self, config, WebServer, DEBUG)
        pass

    def getArgParser(self,description:str,version_msg)->ArgumentParser:
        """
        override the default argparser call
        """        
        parser=super().getArgParser(description, version_msg)
        parser.add_argument("-v", "--verbose", action="store_true", help="show verbose output [default: %(default)s]")
        parser.add_argument("-rp", "--root_path",default=WebServer.examples_path(),help="path to pdf files [default: %(default)s]")
        return parser
    

def main(argv:list=None):
    """
    main call
    """
    cmd=NiceTrackCmd()
    exit_code=cmd.cmd_main(argv)
    return exit_code
        
DEBUG = 0
if __name__ == "__main__":
    if DEBUG:
        sys.argv.append("-d")
    sys.exit(main())


"""
Created on 27.08.2023

@author: wf
"""
from fastapi import Header, HTTPException
from starlette.responses import FileResponse, StreamingResponse
import os
import cv2

class VideoStream():
    """
    wrapper to stream a local file as a video
    """

    def __init__(self, video_path):
        self.video_path = video_path
        self.video_size = os.path.getsize(video_path)
        
    def generate_frames(self, start_time: float = 0.0, fps: float = 30.0):
        cap = cv2.VideoCapture(self.video_path)
        
        # Calculate the starting frame based on start_time and fps
        start_frame = int(start_time * fps)
        cap.set(cv2.CAP_PROP_POS_FRAMES, start_frame)
        
        while True:
            ret, frame = cap.read()
            if not ret:
                cap.release()
                break
            _, buffer = cv2.imencode('.jpg', frame)
            frame = buffer.tobytes()
            yield (b'--frame\r\n'
                   b'Content-Type: image/jpeg\r\n\r\n' + frame + b'\r\n\r\n')


    async def video_feed(self, start_time: float = 0.0, fps: float = 30.0):
        frame_gen = self.generate_frames(start_time=start_time, fps=fps)
        return StreamingResponse(frame_gen, media_type="multipart/x-mixed-replace; boundary=frame")

    def parse_range_header(self, range_header: str) -> tuple:
        """
        Parse the Range header to determine the part of the file to stream.
        Returns a tuple of (start_byte, end_byte).
        """
        start_byte, end_byte = None, None
    
        if range_header:
            try:
                byte1, byte2 = range_header.replace("bytes=", "").split("-")
                start_byte = int(byte1)
                end_byte = int(byte2) if byte2 else None
            except ValueError:
                raise HTTPException(status_code=400, detail="Malformed range header")
    
        return start_byte, end_byte
    
    def stream_partial_content(self, start: int, end: int) -> StreamingResponse:
        """
        Stream the content between the start and end bytes.
        """
        def content_generator():
            with open(self.video_path, 'rb') as video:
                video.seek(start)
                while start <= end:
                    chunk_size = min(8192, end - start + 1)
                    buffer = video.read(chunk_size)
                    if not buffer:
                        break
                    start += chunk_size
                    yield buffer
    
        content_length = end - start + 1
        headers = {
            "Content-Range": f"bytes {start}-{end}/{self.video_size}",
            "Accept-Ranges": "bytes",
            "Content-Length": str(content_length),
            "Content-Type": "video/mp4",
        }
    
        return StreamingResponse(content_generator(), status_code=206, headers=headers)
    
    def get_video(self, range_header: str = Header(default=None)):
        start, end = self.parse_range_header(range_header)
    
        if not start and not end:
            # Just serve the whole file if no range
            return FileResponse(self.video_path, media_type="video/mp4")
    
        if start >= self.video_size:
            raise HTTPException(status_code=416, detail="Range not satisfiable")
    
        end = end or self.video_size - 1
        return self.stream_partial_content(start, end)



"""
Created on 2023-08-31

@author: wf
"""
import io
import os
import cv2
from fastapi.responses import StreamingResponse
from fastapi import HTTPException
from nicegui import ui

class VideoStepper:
    """
    Display a video step by step (frame-by-frame).
    """

    def __init__(self, video_path: str=None, root_path: str=None, fps: int = 30):
        """
        Initialize the VideoStepper instance.
        
        Args:
            video_path (str): Path to the video file.
            root_path (str): Root directory path.
            fps (int, optional): Frames per second. Defaults to 30.
        """
        self.cap=None
        self.root_path=root_path
        self.video_path = None
        self.fps = fps
        self.set_video_path(video_path)
        
    def close(self):
        if self.cap is not None:
            self.cap.release()
        
    def set_video_path(self,video_path:str):
        self.close()
        self.video_path=video_path
        if video_path is None or not os.path.exists(video_path):
            self.video_size=0
            self.base_url=None
            self.fps=1
            # dummy image
            self.url="https://picsum.photos/id/28/1024/768"  
            return
        self.video_size = os.path.getsize(video_path)
        if self.root_path:
            if video_path.startswith(self.root_path):
                video_path = video_path.replace(self.root_path, "")
                if video_path.startswith("/"):
                    # replace first slash
                    video_path = video_path.replace("/", "", 1)
                else:
                    video_path = os.path.basename(video_path)
        self.base_url=f"/video_step/{video_path}"   
        self.cap = cv2.VideoCapture(self.video_path)
        self.set_frame_index(0)
        
    def set_frame_index(self,frame_index:int=0):
        self.frame_index=frame_index
        if self.base_url:
            self.url = f"{self.base_url}/{self.frame_index}"
        
    def get_view(self,container):
        with container:
            self.view=ui.interactive_image(
                events=['click', 'mousedown', 'mouseup'], 
                cross=True
            ).bind_source_from(self, 'url').bind_content(self, 'svg_content')
        return self.view
        
    def get_frame(self, frame: int=None):
        """
        Extracts the specified frame from the video.

        Args:
            frame (int): The frame number to extract.

        Returns:
            numpy.ndarray or None: The image of the specified frame or None if the frame doesn't exist.
        """
        if self.cap is None:
            return None
        if frame is None:
            frame=self.frame_index
            
        # Calculate the frame number
        frame_number = frame * self.fps

        # Set the position of the video file capture to the desired frame
        self.cap.set(cv2.CAP_PROP_POS_FRAMES, frame_number)

        ret, frame = self.cap.read()

        if ret:
            return frame
        else:
            return None
   
    def get_image(self, frame: int=None, img_format: str = "jpg") -> bytes:
        """
        Extracts the specified frame from the video and converts it to the desired image format.

        Args:
            frame (int): The frame number to extract.
            img_format (str, optional): Image format, supports "jpg" and "png". Defaults to "jpg".

        Returns:
            bytes or None: The image bytes of the specified frame in the desired format or None if the frame doesn't exist.
        """
        if frame is None:
            frame=self.frame_index
            
        # Check if the provided format is supported
        if img_format not in ["jpg", "png"]:
            raise ValueError("Unsupported image format. Please use 'jpg' or 'png'.")
        
        # Get the frame from the video
        frame_img = self.get_frame(frame)
        
        # If frame exists, encode it to the desired format
        if frame_img is not None:
            retval, buffer = cv2.imencode(f'.{img_format}', frame_img)
            
            if retval:
                return buffer.tobytes()
        return None
    
    async def stream_image(self, frame_index: int = 0, img_format: str = "jpg") -> StreamingResponse:
        """
        Stream an image from the video at the specified frame index in the given image format.
    
        Args:
            frame_index (int, optional): Frame index to retrieve. Defaults to 0.
            img_format (str, optional): Image format, supports "jpg" and "png". Defaults to "jpg".
        
        Raises:
            HTTPException: Raises a 404 exception if the video is not available.
        
        Returns:
            StreamingResponse: The image from the specified frame in the desired format.
        """
        if not self.video_size:
            raise HTTPException(status_code=404, detail=f"Video not available")
        
        image_bytes = self.get_image(frame_index, img_format)
        
        if image_bytes:
            return StreamingResponse(io.BytesIO(image_bytes), media_type=f"image/{img_format}")
        else:
            raise HTTPException(status_code=404, detail="Image not found")


'''
Created on 2023-08-16

@author: wf
'''
import os
from nicetrack.srt import SRT
from ngwidgets.basetest import Basetest
import requests

class Test_SRT(Basetest):
    """
    test SRT class
    """
    
    def setUp(self, debug=False, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)

    def fetch_srt_files_from_github(self, repo_url: str, path: str):
        # Extract user and repo name from the given URL
        parts = repo_url.split("/")
        user, repo = parts[-2], parts[-1]
        
        # Construct the API URL
        api_url = f"https://api.github.com/repos/{user}/{repo}/{path}"
    
        headers = {"Accept": "application/vnd.github.v3+json"}
        
        response = requests.get(api_url, headers=headers)
        response.raise_for_status()
        
        files = response.json()
        
        srt_files = [file["download_url"] for file in files if file["name"].lower().endswith(".srt")]
        return srt_files

    def test_srt(self):
        """
        test some srt samples 
        """
        srt_samples=[("""1
00:00:00,000 --> 00:00:00,033
<font size="28">SrtCnt : 1, DiffTime : 33ms
2023-08-15 09:18:24.589
[iso : 200] [shutter : 1/180.0] [fnum : 170] [ev : 1.3] [ct : 5490] [color_md : default] [focal_len : 240] [dzoom_ratio: 10000, delta:0],[latitude: 48.486375] [longitude: 8.375567] [rel_alt: 0.000 abs_alt: 530.095] </font>

2
00:00:00,033 --> 00:00:00,066
<font size="28">SrtCnt : 2, DiffTime : 33ms
2023-08-15 09:18:24.622
[iso : 200] [shutter : 1/180.0] [fnum : 170] [ev : 1.3] [ct : 5490] [color_md : default] [focal_len : 240] [dzoom_ratio: 10000, delta:0],[latitude: 48.486375] [longitude: 8.375566] [rel_alt: 0.000 abs_alt: 530.095] </font>

3
00:00:00,066 --> 00:00:00,100
<font size="28">SrtCnt : 3, DiffTime : 34ms
2023-08-15 09:18:24.655
[iso : 200] [shutter : 1/180.0] [fnum : 170] [ev : 1.3] [ct : 5490] [color_md : default] [focal_len : 240] [dzoom_ratio: 10000, delta:0],[latitude: 48.486375] [longitude: 8.375566] [rel_alt: 0.000 abs_alt: 530.095] </font>
""",3,48.486375,8.375567,0.00001),("""1
00:00:01,000 --> 00:00:02,000
HOME(149.0251,-20.2532) 2017.08.05 14:11:51
GPS(149.0251,-20.2533,16) BAROMETER:1.9
ISO:100 Shutter:60 EV: Fnum:2.2""",1,-20.2533,16,149.0251)]
        # Load the SRT text using the from_text classmethod
        for _i,srt_sample in enumerate(srt_samples):
            srt_text,count,lat,lon,delta=srt_sample
            srt = SRT.from_text(srt_text)
            self.assertEqual(count,len(srt.subtitles))
            srt.debug=True
            geo_path=srt.as_geopath()
            self.assertEqual(count,len(geo_path.path))
            for tp in geo_path.path:        
                self.assertAlmostEqual(tp.lat, lat, delta=delta)
                self.assertAlmostEqual(tp.lon, lon, delta=delta)
                
    def check_sample(self,srt_file_url:str,hint:str,debug:bool=True):
        response = requests.get(srt_file_url)
        response.raise_for_status()

        srt_text = response.text
        file_name = srt_file_url.split('/')[-1]
        return self.check_text(srt_text,hint,file_name,debug)
        
    def check_text(self,srt_text,hint,file_name,debug):
        srt=SRT.from_text(srt_text)
        geo_path=srt.as_geopath()
        total_distance=geo_path.total_distance()
        
        if debug:
            print(f"{hint}:{len(geo_path.path)}:{total_distance:.3f} km:{file_name}")
        return srt
                
    def test_drone_videos(self):
        """
        Test method to check all SRT files from a local test directory.
        """
        # local test - set you own drone videos path here
        dronevideos_path = "/Volumes/videos/quadrocopter/2023"
        
        if os.path.exists(dronevideos_path):
            index=0
            file_list=sorted(os.listdir(dronevideos_path))
            for filename in file_list:
                if filename.endswith(".SRT") and not filename.startswith("._"):
                    index+=1
                    #print(filename)
                    with open(os.path.join(dronevideos_path, filename), 'r', encoding='utf-8') as file:
                        srt_text = file.read()
                        self.check_text(srt_text, hint=f"{index:2}", file_name=filename, debug=True)
                    
    def test_sample(self):
        sample_url="https://raw.githubusercontent.com/JuanIrache/dji-srt-viewer/master/samples/sample4.SRT"
        srt=self.check_sample(sample_url,"sample4",debug=True)
        debug=True
        for index in range(len(srt.subtitles)):
            d=srt.as_dict(index)
            if debug:
                print(f"{index:3}:{d}")
        #self.assertEqual("2018-08-19 08:00:36",str(d))
        
    def test_samples(self):
        """
        test JuanIrache samples
        """
        debug=self.debug
        debug=True
        repo_urls = [
            "https://github.com/JuanIrache/DJI_SRT_Parser",
            "https://github.com/JuanIrache/dji-srt-viewer"]
        for ir,repo_url in enumerate(repo_urls):
            srt_files = self.fetch_srt_files_from_github(repo_url, "contents/samples")
    
            for i,srt_file_url in enumerate(srt_files):
                t=len(srt_files)
                hint=f"{ir+1}:{i+1:2d}/{t:2d}"
                self.check_sample(srt_file_url,hint,debug)
                


"""
Created on 2023-06-19

@author: wf
"""
from fastapi import HTTPException, Header, Query
from nicetrack.version import Version
from ngwidgets.leaflet import leaflet
from nicetrack.srt import SRT
from nicetrack.geo import GeoPath
from ngwidgets.file_selector import FileSelector
from nicegui import ui, Client
#from nicetrack.video import Video
from nicetrack.video_stream import VideoStream
from nicetrack.video_stepper import VideoStepper
from nicetrack.video_stepper_av import VideoStepperAV
import os

from ngwidgets.input_webserver import InputWebserver
from ngwidgets.webserver import WebserverConfig

class WebServer(InputWebserver):
    """
    WebServer class that manages the server 
    
    """
    
    @classmethod
    def get_config(cls):
        copy_right="(c)2023 Wolfgang Fahl"
        config=WebserverConfig(copy_right=copy_right,version=Version(),default_port=9859)
        return config
  
    def __init__(self):
        """
        Constructs all the necessary attributes for the 
        WebServer object.
        """
        config=WebServer.get_config()
        InputWebserver.__init__(self,config=config)
        self.geo_desc=None
        self.geo_path=None
        self.zoom_level=9
        self.video_stepper=None
        
        @ui.page('/')
        async def home(client: Client):
            return await self.home(client)
            
        @ui.page('/settings')
        async def settings():
            return await self.settings()
            
        @ui.page('/video_play/{video_path:path}')
        async def video_play(video_path: str, range_header: str = Header(None)):
            return await self.video_play(video_path, range_header)
            
        @ui.page('/video_feed/{video_path:path}')    
        async def video_feed(video_path:str,start_time: float = Query(0.0, description="Start timestamp in seconds"), fps: float = Query(30.0, description="Frames per second of the video")):
            return await self.video_feed(video_path,start_time=start_time, fps=fps)

        @ui.page('/video_step/{video_path:path}/{frame_index:int}')
        async def video_step(video_path: str, frame_index:int=0):
            return await self.video_step(video_path, frame_index)
        
    @classmethod
    def examples_path(cls)->str:
        # the root directory (default: examples)
        path = os.path.join(os.path.dirname(__file__), '../nicetrack_examples')
        path = os.path.abspath(path)
        return path
        
    def set_zoom_level(self,zoom_level):
        self.zoom_level=zoom_level
        with self.geo_map as geo_map:
            geo_map.set_zoom_level(zoom_level)
            
    def get_video_stream(self,video_path:str):
        # Check if not local
        if not self.is_local:
            raise HTTPException(status_code=400, detail="Videos only available in local mode of server")
        # Check if the video name exists in the root path
        video_source=os.path.join(self.root_path, video_path)
        if not os.path.exists(video_source):
            raise HTTPException(status_code=404, detail=f"Video {video_source} not found")
        video_stream=VideoStream(video_source)
        return video_stream
    
    async def video_play(self,video_path:str,range_header: str):
        video_stream=self.get_video_stream(video_path)
        stream_response=video_stream.get_video(range_header)
        return stream_response
    
    async def video_feed(self,video_path:str,start_time:float=0.0,fps:int=30.0):
        video_stream=self.get_video_stream(video_path)
        stream_response=await video_stream.video_feed(start_time,fps)
        return stream_response
    
    async def video_step(self,video_path:str,frame_index:int=0):
        if self.video_stepper is None:
            self.video_stepper=VideoStepper(video_path,self.root_path)
        stream_response=await self.video_stepper.stream_image(frame_index)
        return stream_response
 
    def mark_trackpoint_at_index(self,index:int):  
        """
        mark the trackpoint at the given index
        
        Args:
            index(int): the index of the position
        """
        if self.geo_path:
            # get the trackpoint
            self.geo_path.validate_index(index)
            tp=self.geo_path.path[index]
            info=tp.get_info(self.geo_path.nominatim,with_details=False)
            loc=(tp.lat,tp.lon)
            self.trackpoint_desc.content=info
            with self.geo_map as geo_map:
                geo_map.set_location(loc,self.zoom_level)
            if self.video_stepper:
                # @TODO make configurable
                fps=30
                frame_index=self.geo_path.get_video_frame_index(index, fps)
                self.video_stepper.set_frame_index(frame_index)    

    async def render(self, _click_args=None):
        """
        Renders the SRT content

        Args:
            click_args (object): The click event arguments.
        """
        try:
            input_source=self.input
            if input_source.startswith("https://cycle.travel/map/journey/"):
                input_source=input_source.replace("/map/journey","/gpx")+".gpx"
            ui.notify(f"rendering {input_source}")
            geo_text=self.do_read_input(input_source)
            if input_source.lower().endswith(".srt"):
                self.srt=SRT.from_text(geo_text)
                self.geo_path=self.srt.as_geopath()
            elif input_source.lower().endswith(".gpx"):
                self.geo_path=GeoPath.from_gpx(geo_text)
            path_len=len(self.geo_path.path)
            self.time_slider._props['max']=path_len
            self.time_slider.value=0
            file_name=""
            tp_index=self.time_slider.value
            tp=self.geo_path.path[tp_index]
            info=tp.get_info(self.geo_path.nominatim)
            try:
                file_name = self.input.split('/')[-1]
            except BaseException as _bex:
                pass
            desc=f"""{file_name}<br>{info}<br>
{path_len} points
"""
            self.geo_desc.content=desc
            with self.geo_map as geo_map:
                path_2d=self.geo_path.as_tuple_list()
                if len(path_2d)>0:
                    loc=path_2d[0]
                    geo_map.set_location(loc,zoom_level=self.zoom_level)
                    print(f"setting location to {loc}")
                    geo_map.draw_path(path_2d)
            # show render result in log
            #self.log_view.push(render_result.stderr)
        except BaseException as ex:
            self.handle_exception(ex,self.do_trace)  
            
    async def on_play(self):
        """
        play the corresponding video
        """
        if not self.input:
            return
        try:
            if self.input.endswith(".SRT"):
                # pyQT video playing
                video_path=self.input.replace(".SRT",".MP4")
                self.video_stepper.set_video_path(video_path)
                pass
        except BaseException as ex:
            self.handle_exception(ex, self.do_trace)
               
    async def home(self, client:Client):
        """Generates the home page with a map"""
        self.setup_menu()
    
        with ui.element("div").classes("w-full"):
            with ui.splitter() as splitter:
                with splitter.before:
                    self.example_selector=FileSelector(path=self.root_path,handler=self.read_and_optionally_render)
                    self.input_input=ui.input(
                         value=self.input,
                         on_change=self.input_changed).props("size=100")
                    self.tool_button(tooltip="reload",icon="refresh",handler=self.reload_file)    
                    if self.is_local:
                        self.tool_button(tooltip="open",icon="file_open",handler=self.open_file)
                        self.tool_button(tooltip="play",icon="play_circle",handler=self.on_play)
                with splitter.after:
                    self.geo_desc=ui.html("")
                    self.trackpoint_desc=ui.html("")    
            with ui.splitter() as splitter:
                with splitter.before:
                    with leaflet().classes('w-full h-96') as self.geo_map:
                        pass
                with splitter.after as self.video_container:
                    self.video_stepper=VideoStepper(None,self.root_path) 
                    self.video_view=self.video_stepper.get_view(self.video_container)
        slider_props='label-always'
        self.zoom_slider = ui.slider(min=1,max=20,step=1,value=self.zoom_level,on_change=lambda e: self.set_zoom_level(e.value))        .props(slider_props)
        self.time_slider = ui.slider(min=0, max=100, step=1, value=50,on_change=lambda e: self.mark_trackpoint_at_index(e.value))        .props(slider_props)
          
        await self.setup_footer()
        
    def configure_run(self):
        self.allowed_urls=[
            "https://raw.githubusercontent.com/JuanIrache/DJI_SRT_Parser/master/samples/",
            "https://raw.githubusercontent.com/JuanIrache/dji-srt-viewer/master/samples/",
            "https://cycle.travel/gpx/",
            "https://cycle.travel/map/journey/",
            self.examples_path(),
            self.root_path
        ]
  


'''
Created on 2023-06-19

@author: wf
'''
import nicetrack
class Version(object):
    """
    Version handling for nicetrack
    """
    name = "nicetrack"
    version = nicetrack.__version__
    date = '2023-08-16'
    updated = '2023-12-18'
    description = 'Nice 3D Track handling'
    
    authors = 'Wolfgang Fahl'
    
    doc_url="https://wiki.bitplan.com/index.php/nicetrack"
    chat_url="https://github.com/WolfgangFahl/nicetrack/discussions"
    cm_url="https://github.com/WolfgangFahl/nicetrack"

    license = f'''Copyright 2023 contributors. All rights reserved.

  Licensed under the Apache License 2.0
  http://www.apache.org/licenses/LICENSE-2.0

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.'''
    longDescription = f"""{name} version {version}
{description}

  Created by {authors} on {date} last updated {updated}"""


'''
Created on 2023-08-27

@author: wf
'''
import vlc
import platform
from PyQt5 import QtWidgets

class Video:
    """
    Video play wrapper
    """
    def __init__(self,video_source:str):
        # Create a basic vlc instance
        self.instance = vlc.Instance()

        self.media = None

        # Create an empty vlc media player
        self.mediaplayer = self.instance.media_player_new()

        self.video_source=video_source
        # In this widget, the video will be drawn
        if platform.system() == "Darwin": # for MacOS
            self.videoframe = QtWidgets.QMacCocoaViewContainer(0)
        else:
            self.videoframe = QtWidgets.QFrame()
        pass
        # getOpenFileName returns a tuple, so use only the actual file name
        self.media = self.instance.media_new(video_source)

        # Put the media in the media player
        self.mediaplayer.set_media(self.media)

        # Parse the metadata of the file
        self.media.parse()
        # The media player has to be 'connected' to the QFrame (otherwise the
        # video would be displayed in it's own window). This is platform
        # specific, so we must give the ID of the QFrame (or similar object) to
        # vlc. Different platforms have different functions for this
        if platform.system() == "Linux": # for Linux using the X Server
            self.mediaplayer.set_xwindow(int(self.videoframe.winId()))
        elif platform.system() == "Windows": # for Windows
            self.mediaplayer.set_hwnd(int(self.videoframe.winId()))
        elif platform.system() == "Darwin": # for MacOS
            self.mediaplayer.set_nsobject(int(self.videoframe.winId()))

    
    def play(self):
            # Create an instance of the VLC player
            player = vlc.MediaPlayer(self.video_source)
        
            # Play the video
            status=player.play()

            pass


"""
OpenSCAD Examples Scraper
Author: OpenAI's ChatGPT
Date: July 23, 2023

Description:
This script is a web scraper designed to download OpenSCAD example code files from the OpenSCAD examples website.
It uses BeautifulSoup and requests libraries to parse HTML and fetch content. The scraper fetches code content
from the ".html" files using specific "<pre><code data-language="javascript">" structure. The code content is then
saved as separate ".scad" files in a directory structure similar to the HTML structure of the website. The script
utilizes the tqdm library to display a progress bar during the scraping process.

Prompts:
- Write a scraper for https://files.openscad.org/examples/ that extracts the <pre> code sections as files in a
  directory structure similar to the HTML structure.
- Add Google docstrings and type hints to the scraper.
- Wrap the scraper in a class called OpenScadExampleScraper.
- Ensure the code is ready for cut & paste and remove any unnecessary ellipsis ("...").
- The scraper should work with the provided HTML index structure.
- Fix the FileNotFoundError issue when downloading files by ensuring the output directory is set correctly.
- Add a tqdm progress bar to show the scraping progress.
- Include docstrings for each method in the class.
- Include a header with the author, date, and description of the script.
- Add the gist of the prompts that lead to this code in the header.
- The resulting files need to be .scad files
- we want a tqdm progress bar on file level 
- add a result message with total directories files and time the scraping took
"""

import os
import time
import requests
from bs4 import BeautifulSoup
from tqdm import tqdm
from typing import Optional

class OpenScadExampleScraper:
    """
    A web scraper to download OpenSCAD example code files from the website.

    Attributes:
        base_url (str): The base URL of the OpenSCAD examples website.
        output_directory (str): The directory to save the downloaded files.
    """

    def __init__(self):
        """
        Initialize the OpenScadExampleScraper.
        """
        self.base_url: str = "https://files.openscad.org/examples/"
        self.output_directory: str = "openscad_examples"

    def get_soup(self, url: str) -> BeautifulSoup:
        """
        Fetch the BeautifulSoup object for a given URL.

        Args:
            url (str): The URL to fetch and parse.

        Returns:
            BeautifulSoup: The BeautifulSoup object representing the parsed HTML content.
        """
        response = requests.get(url)
        response.raise_for_status()
        return BeautifulSoup(response.content, "html.parser")

    def create_directory_if_not_exists(self, directory: str) -> None:
        """
        Create a directory if it does not exist.

        Args:
            directory (str): The path of the directory to create.
        """
        if not os.path.exists(directory):
            os.makedirs(directory)

    def download_code_file(self, url: str, output_path: str) -> None:
        """
        Download a code file from a URL and save it to the specified output path.

        Args:
            url (str): The URL of the code file to download.
            output_path (str): The path to save the downloaded code file.
        """
        response = requests.get(url)
        response.raise_for_status()
        with open(output_path, "wb") as f:
            f.write(response.content)

    def scrape_subdirectories(self, url: str) -> list:
        """
        Scrape subdirectories from the OpenSCAD examples website and return their names and URLs.

        Args:
            url (str): The URL of the main index page.

        Returns:
            list: A list of tuples containing the subdirectory name and URL.
        """
        soup = self.get_soup(url)
        links = soup.find_all("a")
        subdirectories = []
        for link in links:
            href = link.get("href")
            if href.endswith("/"):  # Subdirectory link
                subdir_url = url + href
                subdir_name = href.rstrip("/")
                subdirectories.append((subdir_name, subdir_url))
        return subdirectories

    def scrape_directory(self, url: str, current_directory: Optional[str] = "") -> int:
        """
        Scrape pre code sections from the OpenSCAD examples website and save them as files.

        Args:
            url (str): The URL of the page to scrape.
            current_directory (str, optional): The current directory path. Defaults to "".

        Returns:
            int: The number of files scraped in the current directory.
        """
        soup = self.get_soup(url)
        links = soup.find_all("a")
        file_links = [link for link in links if link.get("href").endswith(".html")]
        file_links_count = len(file_links)

        file_count = 0

        with tqdm(total=file_links_count, desc=f"Scraping {current_directory}", leave=False) as file_progress:
            for link in file_links:
                href = link.get("href")
                code_url = url + href
                code_content = self.get_code_content(code_url)
                if code_content:
                    file_name = href.split("/")[-1]
                    # Adjust file extension to .scad
                    file_name = file_name.replace(".html", ".scad")
                    file_path = os.path.join(self.output_directory, current_directory, file_name)
                    self.create_directory_if_not_exists(os.path.join(self.output_directory, current_directory))
                    with open(file_path, "w") as f:
                        f.write(code_content.strip())
                    file_count += 1
                file_progress.update(1)

        return file_count

    def get_code_content(self, url: str) -> str:
        """
        Fetch the content of a code file from a given URL.

        Args:
            url (str): The URL of the code file.

        Returns:
            str: The content of the code file.
        """
        response = requests.get(url)
        response.raise_for_status()
        soup = BeautifulSoup(response.content, "html.parser")
        content_div = soup.find("div", id="content")
        code_pre = content_div.find("pre")
        code_content = code_pre.get_text() if code_pre else ""
        return code_content

    def run(self) -> None:
        """
        Run the scraping process.
        """
        index_url = self.base_url
        self.create_directory_if_not_exists(self.output_directory)
        subdirectories = self.scrape_subdirectories(index_url)

        total_directories = len(subdirectories)
        total_files = 0

        start_time = time.time()

        for subdir_name, subdir_url in tqdm(subdirectories, desc="Scraping"):
            self.create_directory_if_not_exists(os.path.join(self.output_directory, subdir_name))
            files_scraped = self.scrape_directory(subdir_url, subdir_name)
            total_files += files_scraped

        end_time = time.time()
        elapsed_time = end_time - start_time

        print("\nScraping Complete!")
        print(f"Total Directories: {total_directories}")
        print(f"Total Files Scraped: {total_files}")
        print(f"Time Taken: {elapsed_time:.2f} seconds")

if __name__ == "__main__":
    scraper = OpenScadExampleScraper()
    scraper.run()



'''
Created on 2023-08-17

@author: wf
'''
from nicetrack.geo import GeoPath
from ngwidgets.basetest import Basetest
import os
from pathlib import Path

class Test_GeoPath(Basetest):
    """
    Test the GeoPath functions
    """
    
    def setUp(self, debug=False, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)

    def test_distance(self):
        """
        test haversine distance calculation
        """
        geo_path = GeoPath()
        geo_path.add_point(52.5200, 13.4050)  # Berlin
        geo_path.add_point(48.8566, 2.3522)   # Paris
        geo_path.add_point(51.5074, -0.1278)  # London
        berlin_paris=geo_path.distance(0, 1)
        berlin_london=geo_path.total_distance()                                                             
        debug=self.debug
        debug=True
        if (debug):
            print(f"Berlin-Paris: {berlin_paris:.0f} km")  
            print(f"Berlin-Paris-London: {berlin_london:.0f} km")
        self.assertAlmostEqual(877,berlin_paris,delta=1)
        self.assertAlmostEqual(1221,berlin_london,delta=1)
        
    def test_reverse_geocoding(self):
        """
        see
        https://github.com/WolfgangFahl/nicetrack/issues/7
        """    
        # avoid forbidden status
        geo_path=GeoPath.from_points((-36.848461, 174.763336)) # Auckland
        if not self.inPublicCI():
            details=geo_path.get_start_location_details()
            debug=self.debug
            debug=True
            if debug:
                print(details)
            self.assertTrue("Auckland" in details)
        
    def test_dms(self):
        """
        test conversion to degrees, minutes/seconds
        """
        geo_path = GeoPath.from_points((25.19683, 55.27395)) # Burj Khalifa
        dms=geo_path.as_dms(0)
        debug=self.debug
        debug=True
        if debug:
            print(dms)
        expected="25° 11' 48.5880'' N", "55° 16' 26.2200'' E"
        self.assertEqual(expected, dms)
        
    def test_google_maps_link(self):
        """
        test getting a google maps link
        """
        geo_path = GeoPath.from_points((25.19683, 55.27395)) # Burj Khalifa
        link=geo_path.as_google_maps_link(0)
        debug=self.debug
        #debug=True
        if debug:
            print(link)
        self.assertTrue("q=25.19683,55.27395" in link)
        
        
    def test_gpx(self):
        """
        test GPX reading with GeoPath class
        """
        # Getting the directory of the current script/file
        current_dir = Path(__file__).parent
    
        # Forming the complete path to the GPX file
        gpx_file_path = os.path.join(current_dir, '..', 'nicetrack_examples', 'gpx', '149759.gpx')
        
        with open(gpx_file_path, 'r') as gpx_file:
            gpx_text=gpx_file.read()
            geo_path=GeoPath.from_gpx(gpx_text)
            debug=True
            track_points=len(geo_path.path)
            if debug:
                print(f"{track_points} track points found")
            self.assertEqual(2334,track_points)
    
        


'''
Created on 16.08.2023

@author: wf
'''
from typing import List, Tuple
import math
import logging
import os
from OSMPythonTools.nominatim import Nominatim
from OSMPythonTools.cachingStrategy import CachingStrategy, JSON
from pathlib import Path
import gpxpy
from dataclasses import dataclass
from datetime import datetime

@dataclass
class Trackpoint:
    """
    a 4D geographic trackpoint
    """
    lat: float
    lon: float
    elevation: float = None
    timestamp: datetime = None
    
    def decimal_to_dms(self, decimal_degree: float):
        degrees = int(decimal_degree)
        minutes = int((decimal_degree - degrees) * 60)
        seconds = (decimal_degree - degrees - minutes/60) * 3600.0
        return degrees, minutes, seconds

    def lat_lon_to_dms_string(self, latitude: float, longitude: float) -> str:
        lat_degree, lat_minute, lat_second = self.decimal_to_dms(latitude)
        lon_degree, lon_minute, lon_second = self.decimal_to_dms(longitude)
        lat_direction = "N" if latitude >= 0 else "S"
        lon_direction = "E" if longitude >= 0 else "W"
        lat_str = f"{abs(lat_degree)}° {abs(lat_minute)}' {abs(lat_second):.4f}'' {lat_direction}"
        lon_str = f"{abs(lon_degree)}° {abs(lon_minute)}' {abs(lon_second):.4f}'' {lon_direction}"
        return lat_str, lon_str
    
    def as_dms(self)->str:
        dms=self.lat_lon_to_dms_string(self.lat, self.lon)
        return dms
    
    def as_google_maps_link(self)->str:
        link = f"https://maps.google.com/?q={self.lat},{self.lon}"
        return link
    
    def as_google_maps_anchor(self)->str:
        google_maps_link=self.as_google_maps_link()
        lat_str,lon_str=self.as_dms()
        a=f"<a href='{google_maps_link}' title='google maps' target='_blank'>{lat_str}{lon_str}</a>"
        return a
    
    def get_details(self,nominatim):
        """
        get the location details via nominatim
        """
        result = nominatim.query(f'{self.lat}, {self.lon}').toJSON()
        location=result[0]
        details = location.get('display_name', "Location not found")
        return details
    
    def get_info(self,nominatim,with_details:bool=True)->str:
        if with_details:
            details=self.get_details(nominatim)
        else:
            details=""
        geo_date_html = "" if self.timestamp is None else f"{self.timestamp}<br>\n"
        google_maps_a=self.as_google_maps_anchor()
        info=f"{geo_date_html}{details}{google_maps_a}"
        return info

class GeoPath:
    """
    a 3D Path of geographic coordinates in lat/lon notation
    """
    def __init__(self, name: str=None, cacheDir: str=None):
        self.name = name
        self.path: List[Trackpoint] = []
        if cacheDir is None:
            home = str(Path.home())
            cacheDir = f"{home}/.nominatim"
        self.cacheDir = cacheDir
        if not os.path.exists(self.cacheDir):
            os.makedirs(self.cacheDir)
        logging.getLogger('OSMPythonTools').setLevel(logging.ERROR)
        CachingStrategy.use(JSON, cacheDir=self.cacheDir)
        self.nominatim = Nominatim()

    @classmethod
    def from_points(cls, *points) -> "GeoPath":
        """
        get a geopath for the given points
        """
        geo_path = cls()
        for point in points:
            lat, lon = point
            geo_path.add_point(lat, lon)
        return geo_path

    @classmethod
    def from_gpx(cls, gpx_data: str) -> "GeoPath":
        geo_path = cls()
        gpx = gpxpy.parse(gpx_data)
        for track in gpx.tracks:
            for segment in track.segments:
                for point in segment.points:
                    geo_path.add_point(point.latitude, point.longitude, point.elevation, point.time)
        return geo_path

    def add_point(self, lat: float, lon: float, elevation: float = None, timestamp: datetime = None) -> None:
        self.path.append(Trackpoint(lat, lon, elevation, timestamp))

    def get_path(self) -> List[Trackpoint]:
        return self.path
    
    def as_tuple_list(self) -> List[Tuple[float, float]]:
        return [(point.lat, point.lon) for point in self.path]

    def haversine_distance(self, point1: Trackpoint, point2: Trackpoint) -> float:
        R = 6371.0
        lat1 = math.radians(point1.lat)
        lon1 = math.radians(point1.lon)
        lat2 = math.radians(point2.lat)
        lon2 = math.radians(point2.lon)
        dlat = lat2 - lat1
        dlon = lon2 - lon1
        a = math.sin(dlat / 2)**2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2)**2
        c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
        distance = R * c
        return distance

    def validate_index(self, index: int):
        max_index = len(self.path) - 1
        if index < 0 or index > max_index:
            raise ValueError(f"Invalid index {index}: valid range is 0-{max_index}")

    def distance(self, index1: int, index2: int) -> float:
        self.validate_index(index1)
        self.validate_index(index2)
        index1, index2 = min(index1, index2), max(index1, index2)
        total_distance = 0
        for i in range(index1, index2):
            total_distance += self.haversine_distance(self.path[i], self.path[i + 1])
        return total_distance

    def total_distance(self) -> float:
        total_distance = 0
        if len(self.path) > 1:
            total_distance = self.distance(0, len(self.path) - 1)
        return total_distance


    def as_dms(self, index: int) -> Tuple[str, str]:
        self.validate_index(index)
        tp = self.path[index]
        dms = tp.as_dms()
        return dms

    def as_google_maps_link(self, index: int) -> str:
        self.validate_index(index)
        tp=self.path[index]
        link=tp.as_google_maps_link()
        return link

    def get_start_location_details(self) -> str:
        if len(self.path) < 1:
            return "No points in path"
        tp=self.path[0]
        details=tp.get_details(self.nominatim)
        return details
    
    def get_video_frame_index(self, index: int, fps: int) -> int:
        """
        Return a video frame index based on a geo_path index and fps parameter using the timestamps.
        
        Args:
        - geo_index (int): The specific index on the GeoPath.
        - fps (int): Frames per second of the video.
        
        Returns:
        - int: Corresponding frame index.
        """
        self.validate_index(index)
        
        start_time = self.path[0].timestamp
        target_time = self.path[index].timestamp

        # Calculate time difference in seconds
        delta_seconds = (target_time - start_time).total_seconds()

        # Calculate frame index based on the time difference and fps
        frame_index = int(delta_seconds * fps)
        
        return frame_index


'''
Created on 2023-08-16

@author: wf
'''
from datetime import datetime
from nicetrack.geo import GeoPath
from typing import Optional
import json
import pysrt
import re
import sys
import traceback


class SRT:
    """
    Class to represent and manipulate SRT (SubRip Subtitle) data.
    
    Attributes:
        subtitles (list): List of parsed subtitles.
    """

    def __init__(self, subtitles, patterns=None, debug:bool=False):
        """
        Initializes the SRT object with the given subtitles and extraction patterns.
        """
        self.subtitles = subtitles
        self.debug = debug
        # Default patterns
        self.patterns = {
            'key_value': r'\[(\w+)\s*:\s*([\d.-]+)\]',
            'double_key_value': r'\[(\w+)\s*:\s*([\d.-]+)\s+(\w+)\s*:\s*([\d.-]+)\]',
            'bracketed_value': r'(\w+)\(([\d.-]+),([\d.-]+)(?:,([\d.-]+))?\)'
        }

        if patterns:
            self.patterns.update(patterns)

    @classmethod
    def from_text(cls, text):
        """
        Class method to create an SRT object from a raw text string.
        """
        subtitles = pysrt.from_string(text)
        return cls(subtitles)
    
    def handle_exception(self, e: BaseException, trace: Optional[bool] = False):
        """Handles an exception by creating an error message.

        Args:
            e (BaseException): The exception to handle.
            trace (bool, optional): Whether to include the traceback in the error message. Default is False.
        """
        if trace:
            error_msg = str(e) + "\n" + traceback.format_exc()
        else:
            error_msg = str(e)
        if self.debug:
            print(error_msg,file=sys.stderr)
    
    def as_dict(self,index: int) -> dict:
        """
        convert the subtitle at the given index to a dict
        """
        if index < len(self.subtitles):
            s = self.subtitles[index].text
        if "<font" in s:
            d=self.as_dji_dict(s)
        else:
            d=self.as_srt_dict(s)
        d=self.as_unified_dict(d)
        return d
    
    def as_unified_dict(self, data_dict: dict) -> dict:
        unified_dict = {}
    
        # Key look-up map along with their respective type-casting functions
        lookup = {
            "lat": (["latitude", "gps_latitude"], float),
            "lon": (["longitude", "gps_longitude"], float),
            "elevation": (["abs_alt", "barometer"], float),
            "timestamp": (["timestamp"], None)
        }
    
        for unified_key, (possible_keys, type_case) in lookup.items():
            for key in possible_keys:
                if key in data_dict:
                    value = data_dict[key]
                    if type_case:
                        if value:
                            value = type_case(value)
                    unified_dict[unified_key] = value
                    break  # Break once we've found the first matching key
    
        return unified_dict



    def as_dji_dict(self,s: str) -> dict:
        """
        Convert a provided string with timestamps and metadata into a dictionary.
        
        Parameters:
            s (str): Input string with timestamps and metadata.
            
        Returns:
            dict: Dictionary with extracted data.
        
        Example:
        --------
        s = '''00:00:00,000 --> 00:00:00,033
        <font size="28">SrtCnt : 1, DiffTime : 33ms
        2023-08-15 09:18:24.589
        [iso : 200] [shutter : 1/180.0] [fnum : 170] [ev : 1.3] [ct : 5490] 
        [color_md : default] [focal_len : 240] [dzoom_ratio: 10000, delta:0],
        [latitude: 48.486375] [longitude: 8.375567] [rel_alt: 0.000 abs_alt: 530.095] </font>'''
        
        result = as_dji_dict(s)
        print(result)
        """
    
        # Extract SrtCnt and DiffTime
        srtcnt_pattern = r"SrtCnt : (\d+)"
        difftime_pattern = r"DiffTime : (\d+ms)"
        
        srtcnt = re.search(srtcnt_pattern, s).group(1) if re.search(srtcnt_pattern, s) else None
        difftime = re.search(difftime_pattern, s).group(1) if re.search(difftime_pattern, s) else None
    
        # Extract the main timestamp
        main_timestamp_pattern = r"(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{3})"
        main_timestamp = re.search(main_timestamp_pattern, s).group(1) if re.search(main_timestamp_pattern, s) else None
    
        # Extract the start and end timestamps
        timestamp_pattern = r"(\d{2}:\d{2}:\d{2},\d{3}) --> (\d{2}:\d{2}:\d{2},\d{3})"
        timestamp_match = re.search(timestamp_pattern, s)
        start_time = timestamp_match.group(1) if timestamp_match else None
        end_time = timestamp_match.group(2) if timestamp_match else None
    
        # Extract metadata enclosed within square brackets
        metadata_pattern = r"\[(.*?)\]"
        metadata_matches = re.findall(metadata_pattern, s)
        metadata_dict = {}
        for match in metadata_matches:
            # Handling for rel_alt and abs_alt specifically
            if "rel_alt" in match and "abs_alt" in match:
                rel_alt_match = re.search(r"rel_alt:\s*([-+]?\d*\.\d+|\d+)", match)
                abs_alt_match = re.search(r"abs_alt:\s*([-+]?\d*\.\d+|\d+)", match)
                if rel_alt_match:
                    metadata_dict['rel_alt'] = float(rel_alt_match.group(1))
                if abs_alt_match:
                    metadata_dict['abs_alt'] = float(abs_alt_match.group(1))
                continue
    
            # Splitting the rest of the metadata
            metadata_split = match.split(":")
            if len(metadata_split) >= 2:
                key = metadata_split[0].strip()
                if len(metadata_split) == 2:
                    value = metadata_split[1].strip()
                else:
                    value = ":".join(metadata_split[1:]).strip()
                metadata_dict[key] = value
    
        # Adding additional metadata to the dictionary
        metadata_dict['start_time'] = start_time
        metadata_dict['end_time'] = end_time
        metadata_dict['SrtCnt'] = int(srtcnt) if srtcnt else None
        metadata_dict['DiffTime'] = difftime
        metadata_dict['timestamp_str'] = main_timestamp
        metadata_dict['timestamp'] = datetime.strptime(main_timestamp, "%Y-%m-%d %H:%M:%S.%f") if main_timestamp else None
        
        return metadata_dict

   
    def as_srt_dict(self,s:str) -> dict:
        """
        Convert SRT format to dict.
        
        Parameters:
            s (str): Input string in SRT format containing metadata.
        
        Returns:
            dict: Dictionary representation of the input metadata.
        
        Example:
        --------
        s = '''1
        00:00:01,000 --> 00:00:02,000
        HOME(149.0251,-20.2532) 2017.08.05 14:11:51
        GPS(149.0251,-20.2533,16) BAROMETER:1.9
        ISO:100 Shutter:60 EV: Fnum:2.2'''
        
        result = as_srt_dict(s)
        print(result)
        """
        # Regular expression to capture key-value pairs
        pattern = r'(\w+):\s*([\d.]+)'
        
        # Find all key-value pairs
        kv_matches = re.findall(pattern, s)
        
        # Extract HOME and GPS data
        home_match = re.search(r'HOME\(([-\d.]+),([-\d.]+)\)', s)
        gps_match = re.search(r'GPS\(([-\d.]+),([-\d.]+),([\d.]+)\)', s)
        
        # Extract timestamp
        date_match = re.search(r'(\d{4}.\d{2}.\d{2} \d{2}:\d{2}:\d{2})', s)
        
        # Build the initial data dictionary
        data = {
            'home_latitude': float(home_match.group(2)) if home_match else None,
            'home_longitude': float(home_match.group(1)) if home_match else None,
            'gps_latitude': float(gps_match.group(2)) if gps_match else None,
            'gps_longitude': float(gps_match.group(1)) if gps_match else None,
            'gps_3': float(gps_match.group(3)) if gps_match else None,
            'timestamp': datetime.strptime(date_match.group(1), "%Y.%m.%d %H:%M:%S") if date_match else None,
        }
        
        # Append additional key-value pairs to the data dictionary
        for key, value in kv_matches:
            # Convert the string value to its appropriate type (integer or float)
            try:
                if '.' in value:
                    value = float(value)
                else:
                    value = int(value)
            except ValueError:
                pass  # keep it as string if conversion fails
            
            # Add to the data dictionary
            data[key.lower()] = value
        
        return data

    def as_geopath(self) -> GeoPath:
        """
        Converts the SRT object into a GeoPath object by extracting latitudes and longitudes.
        """
        geo_path = GeoPath()
        
        for index in range(len(self.subtitles)):
            try:
                d=self.as_dict(index)
                if self.debug:
                    print(json.dumps(d,indent=2,default=str))
                lat=d.get("lat")
                lon=d.get("lon")
                elevation=d.get("elevation")
                timestamp=d.get("timestamp")  
                if lat and lon:  
                    geo_path.add_point(lat, lon, elevation, timestamp)
            except BaseException as ex:
                self.handle_exception(ex, trace=self.debug)

        return geo_path

        



'''
Created on 2023-01

https://github.com/zauberzeug/nicegui/blob/main/examples/map/main.py

@author: rodja
'''
from ngwidgets import leaflet.leaflet  # this module wraps the JavaScript lib leafletjs.com into an easy-to-use NiceGUI element

from nicegui import Client, ui

class MapExample:
    """
    Leaflet based map example
    """

    def __init__(self):
        self.zoom_level=9
        self.locations = {
            (48.486375, 8.375567): 'Baiersbronn',
            (52.5200, 13.4049): 'Berlin',
            (52.37487,9.74168): 'Hannover',
            (40.7306, -74.0060): 'New York',
            (39.9042, 116.4074): 'Beijing',
            (35.6895, 139.6917): 'Tokyo',
        }
        
        self.zoom_levels = {
            9:9,
            10:10,
            11:11,
            12:12,
            13:13,
            14:14,
            15:15,
            16:16,
            17:17,
        }
        
        @ui.page('/')
        async def main_page(client: Client):
            await self.show_map(client)
            
    def on_change_zoom_level(self,_ce):
        self.zoom_level=_ce.value
        self.map.set_zoom_level(self.zoom_level)
        pass
    
    async def draw_path(self,_ce):
        """
        """
        with self.map as geo_map:
            locs=list(self.locations.keys())
            path=[locs[1],locs[2]]
            geo_map.draw_path(path)
            
            
    async def show_map(self,client:Client):
        self.map = leaflet().classes('w-full h-96')
        selection = ui.select(self.locations, on_change=lambda e: self.map.set_location(e.value,zoom_level=self.zoom_level)).classes('w-40')
        zoom_selection=ui.select(self.zoom_levels, on_change=self.on_change_zoom_level).classes('w-40')
        await client.connected()  # wait for websocket connection
        selection.set_value(next(iter(self.locations)))  # trigger map.set_location with first location in selection
        zoom_selection.set_value(self.zoom_level)
        ui.button("Path",on_click=self.draw_path)
    
    def run(self):    
        ui.run(title="map example",reload=False)
    
if __name__ == "__main__":
    map_example=MapExample()
    map_example.run()



import io
import os
import av
from fastapi.responses import StreamingResponse
from fastapi import HTTPException
from nicegui import ui

class VideoStepperAV:
    """
    Display a video step by step (frame-by-frame) using PyAV for video decoding.
    """

    def __init__(self, video_path: str=None, root_path: str=None, fps: int = 30):
        self.container = None
        self.root_path = root_path
        self.video_path = None
        self.fps = fps
        self.set_video_path(video_path)

    def close(self):
        if self.container is not None:
            self.container.close()

    def set_video_path(self, video_path: str):
        self.close()
        self.video_path = video_path
        if video_path is None or not os.path.exists(video_path):
            self.video_size = 0
            self.base_url = None
            self.fps = 1
            # dummy image
            self.url = "https://picsum.photos/id/28/1024/768"
            return
        self.video_size = os.path.getsize(video_path)
        if self.root_path:
            if video_path.startswith(self.root_path):
                video_path = video_path.replace(self.root_path, "")
                if video_path.startswith("/"):
                    video_path = video_path.replace("/", "", 1)
                else:
                    video_path = os.path.basename(video_path)
        self.base_url = f"/video_step/{video_path}"
        self.container = av.open(self.video_path)
        self.set_frame_index(0)

    def set_frame_index(self, frame_index: int = 0):
        self.frame_index = frame_index
        if self.base_url:
            self.url = f"{self.base_url}/{self.frame_index}"

    def get_view(self, container):
        with container:
            self.view = ui.interactive_image(
                events=['click', 'mousedown', 'mouseup'],
                cross=True
            ).bind_source_from(self, 'url').bind_content(self, 'svg_content')
        return self.view

    def get_frame(self, frame_index: int = None):
        if self.container is None:
            return None
        if frame_index is None:
            frame_index = self.frame_index

        # Seek to the desired frame
        video_stream = self.container.streams.video[0]
        frame_num = frame_index
        framerate = video_stream.average_rate
        time_base = video_stream.time_base
        sec = int(frame_num / framerate)
        self.container.seek(sec * 1000000, whence='time', backward=True)
        
        # Get the next available frame
        for _ in range(sec, frame_num):
            frame = next(self.container.decode(video=0))
        return frame.to_image()

    def get_image(self, frame_index: int = None, img_format: str = "jpg") -> bytes:
        if frame_index is None:
            frame_index = self.frame_index

        # Check if the provided format is supported
        if img_format not in ["jpg", "png"]:
            raise ValueError("Unsupported image format. Please use 'jpg' or 'png'.")

        # Get the frame from the video
        frame_img = self.get_frame(frame_index)

        # If frame exists, encode it to the desired format
        if frame_img is not None:
            buffer = io.BytesIO()
            img_format = "jpeg" if img_format == "jpg" else img_format  # Handle 'jpg' -> 'jpeg'
            frame_img.save(buffer, format=img_format)
            return buffer.getvalue()
        return None

    async def stream_image(self, frame_index: int = 0, img_format: str = "jpg") -> StreamingResponse:
        if not self.video_size:
            raise HTTPException(status_code=404, detail=f"Video not available")

        image_bytes = self.get_image(frame_index, img_format)

        if image_bytes:
            return StreamingResponse(io.BytesIO(image_bytes), media_type=f"image/{img_format}")
        else:
            raise HTTPException(status_code=404, detail="Image not found")



'''
Created on 2023-08-19

@author: wf
'''
import os
from pathlib import Path
import gpxpy
from ngwidgets.basetest import Basetest

class Test_GPX(Basetest):
    """
    test GPX  handling
    """

    def test_read_gpx(self):
        """
        Test reading a GPX file
        """
        # Getting the directory of the current script/file
        current_dir = Path(__file__).parent
    
        # Forming the complete path to the GPX file
        gpx_file_path = os.path.join(current_dir, '..', 'nicetrack_examples', 'gpx', '149759.gpx')
        
        with open(gpx_file_path, 'r') as gpx_file:
            gpx = gpxpy.parse(gpx_file)
            debug=self.debug
            debug=True
            track_count=len(gpx.tracks)
            # Count the total number of points across all tracks
            total_points = sum(len(segment.points) for track in gpx.tracks for segment in track.segments)
            
            if debug:
                print(f"Found {track_count} GPX tracks")
                print(f"Total number of points across all tracks: {total_points}")
            
            # Check if the GPX file contains a single track
            self.assertEqual(1, track_count, "Expected 1 track in the GPX file")
            expected_point_count=2334
            # check the point count
            self.assertEqual(expected_point_count, total_points, f"Expected {expected_point_count} points, but found {total_points}")



#! /usr/bin/env python3
#
# PyQt5 example for VLC Python bindings
# Copyright (C) 2009-2010 the VideoLAN team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
#
"""
A simple example for VLC python bindings using PyQt5.

Author: Saveliy Yusufov, Columbia University, sy2685@columbia.edu
Date: 25 December 2018
"""

import platform
import os
import sys

from PyQt5 import QtWidgets, QtGui, QtCore
import vlc

class Player(QtWidgets.QMainWindow):
    """A simple Media Player using VLC and Qt
    """

    def __init__(self, master=None):
        QtWidgets.QMainWindow.__init__(self, master)
        self.setWindowTitle("Media Player")

        # Create a basic vlc instance
        self.instance = vlc.Instance()

        self.media = None

        # Create an empty vlc media player
        self.mediaplayer = self.instance.media_player_new()

        self.create_ui()
        self.is_paused = False

    def create_ui(self):
        """Set up the user interface, signals & slots
        """
        self.widget = QtWidgets.QWidget(self)
        self.setCentralWidget(self.widget)

        # In this widget, the video will be drawn
        if platform.system() == "Darwin": # for MacOS
            self.videoframe = QtWidgets.QMacCocoaViewContainer(0)
        else:
            self.videoframe = QtWidgets.QFrame()

        self.palette = self.videoframe.palette()
        self.palette.setColor(QtGui.QPalette.Window, QtGui.QColor(0, 0, 0))
        self.videoframe.setPalette(self.palette)
        self.videoframe.setAutoFillBackground(True)

        self.positionslider = QtWidgets.QSlider(QtCore.Qt.Horizontal, self)
        self.positionslider.setToolTip("Position")
        self.positionslider.setMaximum(1000)
        self.positionslider.sliderMoved.connect(self.set_position)
        self.positionslider.sliderPressed.connect(self.set_position)

        self.hbuttonbox = QtWidgets.QHBoxLayout()
        self.playbutton = QtWidgets.QPushButton("Play")
        self.hbuttonbox.addWidget(self.playbutton)
        self.playbutton.clicked.connect(self.play_pause)

        self.stopbutton = QtWidgets.QPushButton("Stop")
        self.hbuttonbox.addWidget(self.stopbutton)
        self.stopbutton.clicked.connect(self.stop)

        self.hbuttonbox.addStretch(1)
        self.volumeslider = QtWidgets.QSlider(QtCore.Qt.Horizontal, self)
        self.volumeslider.setMaximum(100)
        self.volumeslider.setValue(self.mediaplayer.audio_get_volume())
        self.volumeslider.setToolTip("Volume")
        self.hbuttonbox.addWidget(self.volumeslider)
        self.volumeslider.valueChanged.connect(self.set_volume)

        self.vboxlayout = QtWidgets.QVBoxLayout()
        self.vboxlayout.addWidget(self.videoframe)
        self.vboxlayout.addWidget(self.positionslider)
        self.vboxlayout.addLayout(self.hbuttonbox)

        self.widget.setLayout(self.vboxlayout)

        menu_bar = self.menuBar()

        # File menu
        file_menu = menu_bar.addMenu("File")

        # Add actions to file menu
        open_action = QtWidgets.QAction("Load Video", self)
        close_action = QtWidgets.QAction("Close App", self)
        file_menu.addAction(open_action)
        file_menu.addAction(close_action)

        open_action.triggered.connect(self.open_file)
        close_action.triggered.connect(sys.exit)

        self.timer = QtCore.QTimer(self)
        self.timer.setInterval(100)
        self.timer.timeout.connect(self.update_ui)

    def play_pause(self):
        """Toggle play/pause status
        """
        if self.mediaplayer.is_playing():
            self.mediaplayer.pause()
            self.playbutton.setText("Play")
            self.is_paused = True
            self.timer.stop()
        else:
            if self.mediaplayer.play() == -1:
                self.open_file()
                return

            self.mediaplayer.play()
            self.playbutton.setText("Pause")
            self.timer.start()
            self.is_paused = False

    def stop(self):
        """Stop player
        """
        self.mediaplayer.stop()
        self.playbutton.setText("Play")

    def open_file(self):
        """Open a media file in a MediaPlayer
        """

        dialog_txt = "Choose Media File"
        filename = QtWidgets.QFileDialog.getOpenFileName(self, dialog_txt, os.path.expanduser('~'))
        if not filename:
            return

        # getOpenFileName returns a tuple, so use only the actual file name
        self.media = self.instance.media_new(filename[0])

        # Put the media in the media player
        self.mediaplayer.set_media(self.media)

        # Parse the metadata of the file
        self.media.parse()

        # Set the title of the track as window title
        self.setWindowTitle(self.media.get_meta(0))

        # The media player has to be 'connected' to the QFrame (otherwise the
        # video would be displayed in it's own window). This is platform
        # specific, so we must give the ID of the QFrame (or similar object) to
        # vlc. Different platforms have different functions for this
        if platform.system() == "Linux": # for Linux using the X Server
            self.mediaplayer.set_xwindow(int(self.videoframe.winId()))
        elif platform.system() == "Windows": # for Windows
            self.mediaplayer.set_hwnd(int(self.videoframe.winId()))
        elif platform.system() == "Darwin": # for MacOS
            self.mediaplayer.set_nsobject(int(self.videoframe.winId()))

        self.play_pause()

    def set_volume(self, volume):
        """Set the volume
        """
        self.mediaplayer.audio_set_volume(volume)

    def set_position(self):
        """Set the movie position according to the position slider.
        """

        # The vlc MediaPlayer needs a float value between 0 and 1, Qt uses
        # integer variables, so you need a factor; the higher the factor, the
        # more precise are the results (1000 should suffice).

        # Set the media position to where the slider was dragged
        self.timer.stop()
        pos = self.positionslider.value()
        self.mediaplayer.set_position(pos / 1000.0)
        self.timer.start()

    def update_ui(self):
        """Updates the user interface"""

        # Set the slider's position to its corresponding media position
        # Note that the setValue function only takes values of type int,
        # so we must first convert the corresponding media position.
        media_pos = int(self.mediaplayer.get_position() * 1000)
        self.positionslider.setValue(media_pos)

        # No need to call this function if nothing is played
        if not self.mediaplayer.is_playing():
            self.timer.stop()

            # After the video finished, the play button stills shows "Pause",
            # which is not the desired behavior of a media player.
            # This fixes that "bug".
            if not self.is_paused:
                self.stop()

def main():
    """Entry point for our simple vlc player
    """
    app = QtWidgets.QApplication(sys.argv)
    player = Player()
    player.show()
    player.resize(640, 480)
    sys.exit(app.exec_())

if __name__ == "__main__":
    main()



import os
import numpy as np
import cv2
import time
import unittest
from ngwidgets.basetest import Basetest
from nicetrack.video_stepper import VideoStepper  # Modify this import based on where your VideoStepper class is
from nicetrack.video_stepper_av import VideoStepperAV  # Modify this import based on where your AV-based VideoStepper class is

class TestVideoStepper(Basetest):
    """
    Test the VideoStepper class
    """
    
    def setUp(self, debug=False, profile=True):
        Basetest.setUp(self, debug=debug, profile=profile)
        self.video_path = "/Volumes/videos/quadrocopter/2023/DJI_0024.MP4"
        self.opencv_stepper = VideoStepper(self.video_path, root_path="/Volumes/videos/quadrocopter/2023/")
        self.pyav_stepper = VideoStepperAV(self.video_path, root_path="/Volumes/videos/quadrocopter/2023/")

    def _test_get_frame(self, stepper):
        if os.path.exists(self.video_path):
            start = time.time()
            frame = stepper.get_frame(5)  # e.g., get the frame of the 5th second
            end = time.time()

            self.assertIsNotNone(frame, "Frame should not be None")
            if isinstance(frame, np.ndarray):
                self.assertTrue(frame.shape[2] == 3, "Frame should have 3 channels (RGB)")

            time_taken = end - start
            return time_taken

    def _test_get_image(self, stepper):
        if os.path.exists(self.video_path):
            start = time.time()
            image_bytes = stepper.get_image(5, img_format="jpg")
            end = time.time()

            self.assertIsNotNone(image_bytes, "Image bytes should not be None")
            image = cv2.imdecode(np.frombuffer(image_bytes, dtype=np.uint8), cv2.IMREAD_COLOR)
            self.assertTrue(isinstance(image, np.ndarray), "Decoded image should be a numpy ndarray")
            self.assertTrue(image.shape[2] == 3, "Decoded image should have 3 channels (RGB)")

            time_taken = end - start
            return time_taken

    @unittest.skipIf(Basetest.inPublicCI(), "queries unreliable wikidata endpoint")
    def test_compare_timing(self):
        """
        """
        opencv_frame_time = self._test_get_frame(self.opencv_stepper)
        pyav_frame_time = self._test_get_frame(self.pyav_stepper)
        opencv_image_time = self._test_get_image(self.opencv_stepper)
        pyav_image_time = self._test_get_image(self.pyav_stepper)

        frame_time_diff = pyav_frame_time - opencv_frame_time
        image_time_diff = pyav_image_time - opencv_image_time

        frame_time_percentage = (frame_time_diff / opencv_frame_time) * 100
        image_time_percentage = (image_time_diff / opencv_image_time) * 100

        print(f"OpenCV Frame Time: {opencv_frame_time:.4f} seconds")
        print(f"PyAV Frame Time: {pyav_frame_time:.4f} seconds")
        print(f"OpenCV Image Time: {opencv_image_time:.4f} seconds")
        print(f"PyAV Image Time: {pyav_image_time:.4f} seconds")

        print(f"Frame Time Difference: {frame_time_diff:.4f} seconds")
        print(f"Image Time Difference: {image_time_diff:.4f} seconds")

        print(f"Frame Time Relative %: {frame_time_percentage:.2f}%")
        print(f"Image Time Relative %: {image_time_percentage:.2f}%")

    def testVersion(self):
        cv_version = cv2.__version__
        debug = self.debug
        if debug:
            print(cv_version)
        major, minor, _ = [int(x) for x in cv_version.split('.')]

        self.assertTrue(
            (major > 4) or (major == 4 and minor >= 8),
            "OpenCV version should be at least 4.8"
        )
